/******************************************************
 *      memo.h: LBM Interpreter Memory Functions Headers
 ******************************************************/
/***********************************
     (C) Copyright 1992-1994; dit/upm
     Distributed under the conditions stated in the
     TOPO General Public License (see file LICENSE)
 ***********************************
 $Log: memo.h,v $
 * Revision 1.3  1994/10/17  16:41:52  lotos
 * cosmetics
 *
 * Revision 1.2  1993/10/18  17:53:22  lotos
 * cleanup interface
 *
 * Revision 1.1  1993/10/16  10:55:19  lotos
 * Initial revision
 *
 ************************************/

/* From memo.c */
char     *emalloc ();

spe      new_spe ();

void     free_environ ();

krnlt    new_kt ();
void     free_kt ();

cond     new_cond ();
void     free_cond ();

CNDlist  new_CNDlist ();
void     free_CNDlist ();

exper    new_exper ();
void     free_exper ();

exper    *new_exper_array ();
void     free_exper_array ();

soffert  *new_soffert ();
void     free_soffert ();
void     free_sset_off ();

rlbset   *new_rlbset ();
void     free_rlbset ();
